unit liteWeightDownloadFile;

// version .6 beta
//copyright Greg Lorriman 2001
//free for any use.
//If you modify this component do not publish without notifying greg@lorriman.com
//and commenting your changes
//History
//.6 problem comments, very minor bug fix, warnings cleanup
//.3 beta added OnDownLoading Event
{

WARNING : see the problems section below

USAGE : set the properties "url" and "destinationFolder" and call execute.

Fill the OnProgress event if you wish to support a progressbar, status messages
or user cancellation (via the "cancel" parameter or by raising an exception,
remember to use application.processmessages).

OnDownloading is for your convenience. It allows you to more easily mimic IE's
downloading dialog mechanism. It is triggered once only at the beginning of the
download when downloading has commenced successfully (first part of the file has
been transferred). It is therefore not equivelant to OnBegin (which just means
that ports are being opened etc). This event gives you the file size, if
available, and the chance to cancel. It is actually part of the "progress"
event, and is triggered the first time a status code of "5" comes through.
OnProgress will also be triggered. You can cancel by setting the "cancel"
parameter to true or raising an exception (in either event). This component does
not attempt to mimic IE's "where do you want to put it?" behaviour; you must
organise that yourself but this event does help.

If something goes wrong or the user cancels then OnEnd will tell you what
happended. You can decide for yourself whether the error merits an exception.

Execute will return "true" if the file was successfully downloaded, false
otherwise.

If you raise an exception in an event the exception will come back through the
"Execute" method, as one would expect.

Normally the downloader puts the item, using it;s original file name (as taken
from the url), in a folder indicated by DestinationFolder. You can use FilePath
in order to down load to a filename of your choice, in which case
DestinationFolder is ignored.

PROBLEMS :

1) with default IE settings a second download of the target url always fetches
from cache totally ignoring an open internet connection. Regrettably the IE
setting in question appears to have higher prority than any flag that can be
used in conjunction with the underlying api call used in this component.
Depending on your use of this component you may need to adjust this
setting in IE(5) :

"Tools->Internet Options->General->Temporary Internet Files->Settings->Every visit to the page"

2)There is a way for other programs to hook into the download mechanism used by
this component. The result is that these programs take over the download and
can make a mess of things (saving the file somewhere else, for example). The
program "Download Accelerator Plus" does this via it's "intergrate with IE"
setting. The registry keys in question are probably located in

HKEY_CLASSES_ROOT\PROTOCOLS\Name-Space Handler\
HKEY_LOCAL_MACHINE\SOFTWARE\SpeedBit\Download Accelerator   -> "Browser Intergration"
HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\Browser Helper Objects

and I've seen other keys that may have a role to play. If you are particluarly
keen to resolve this problem more statisfactorily then use of regmon and
filemon for http://www.sysinternals.com, and apimonitor from
http://www.rohitab.com/apimonitor/, could come in handy. Also look for "Browser
Helper Objects" on msdn.

I don't recommend that you fiddle with these settings, rather warn your users.
Gozilla and others are likely to pose similar problems.

}

interface

uses classes, urlmon, activeX, windows, sysutils;

type

   TDownloadStatus = (dsNULL, dsFindingresource, dsConnecting, dsRedirecting,
      dsBegindownloaddata, dsDownloadingdata, dsEnddownloaddata,
      dsBegindownloadcomponents, dsInstallingcomponents,
      dsEnddownloadcomponents, dsUsingcachedcopy, dsSendingrequest,
      dsClassidavailable, dsMimetypeavailable, dsCachefilenameavailable,
      dsBeginsyncoperation, dsEndsyncoperation, dsBeginuploaddata,
      dsUploadingdata, dsEnduploadingdata, dsProtocolclassid, dsEncoding,
      dsVerfiedmimetypeavailable, dsClassinstalllocation, dsDecoding,
      dsLoadingmimehandler, dsContentdispositionattach, dsFilterreportmimetype,
      dsClsidcaninstantiate, dsIunknownavailable, dsDirectbind, dsRawmimetype,
      dsProxydetecting, dsAcceptranges);

const
   //hmmm, source formatter didn;t do a very good job.
   StatusStrArray: array[low(TDownloadStatus)..high(TDownloadStatus)] of string
   = (
      'NULL', 'Finding resource', 'Connecting', 'Redirecting',
      'Begin download data', 'Downloading data',
      'End download data', 'Begin download components', 'Installing components',
      'End download components', 'Using cached copy', 'Sending request',
      'Classid available',
      'Mime type available', 'Cache filename available', 'Begin sync operation',
      'End sync operation',
      'Begin upload data', 'Uploading data', 'End uploading data',
      'Protocol classid', 'Encoding',
      'Verfied mime type available', 'Class install location', 'Decoding',
      'Loading mime handler',
      'Content disposition attach', 'Filter report mime type',
      'Clsid can instantiate',
      'Iunknown available', 'Direct bind', 'Raw mime type', 'Proxy detecting',
      'Accept ranges');

type

   TLWOnDownloading = procedure(sender: Tobject; size: integer; var cancel:
      boolean) of object;
   TLWDOnProgress = procedure(sender: Tobject; position, max: integer; status:
      TDownLoadStatus; statusStr, extraInfoStr: string; var cancel: boolean) of
      object;
   TOnDownloadEnded = procedure(sender: Tobject; aborted, Error: boolean;
      errorCode: HResult; ErrorStr: string) of object;

   TgfglLiteWeightDownLoader = class(TComponent)

   private

      FUrl: string;
      FOnProgress: TLWDOnProgress;
      FOnBegin: TNotifyEvent;
      FOnLowResources: TNotifyEvent;
      FOnEnd: TOnDownloadEnded;
      FDestinationFolder: string;
//      FTriggeredDownloadingEvent: boolean;
      FLWOnDownloading: TLWOnDownloading;
      FFilePath: string;
      procedure SetUrl(const Value: string);
      procedure SetFilePath(const Value: string);

   protected

      procedure doLowREsources; virtual;
      procedure doBegin; virtual;
      procedure doEnded(aborted, error: boolean; errorCode: HResult; statusStr:
         string); virtual;
      procedure doProgress(position, max: integer; status: TDownLoadStatus;
         statusStr, extraInfoStr: string; var cancel: boolean); virtual;
      procedure doDownloading(size: integer; var cancel: boolean);

   public

      function Execute: boolean; virtual;

   published

      //where to get data
      property Url: string read FUrl write SetUrl;
      //where to put it
      property DestinationFolder: string read FDestinationFolder write
      FDestinationFolder;
      //filepath overrides destinationFolder and the filename in the url
      property FilePath: string read FFilePath write SetFilePath;
      //process has begun - simple notification
      property OnBegin: TNotifyEvent read FOnBegin write FOnBegin;
      //is able to download (in fact is downloading like IE download dialog)
      //like IE you oppotunity to cancel (cancel:=true or exception);
      //only triggered once (at begining of download)
      property OnDownloading: TLWOnDownloading read FLWOnDownloading write
      FLWOnDownloading;
      //download progressing, also triggered when OnDownloading is triggered. Cancel=false by default
      property OnProgress: TLWDOnProgress read FOnProgress write FOnProgress;
      //completed : indicates success through Error parameter (file probably gets removed by windows)
      property OnEnd: TOnDownloadEnded read FOnEnd write FOnEnd;
//not implemented by MS yet
//      property OnLowResources: TNotifyEvent read FOnLowResources write FOnLowResources;

   end;

function extractObjectName(url: string): string;

procedure Register;

implementation

uses httpApp, filectrl;

type
   TUrlCallBack = class(TInterfacedObject, IBindStatusCallBack)
   private

      FErrorCode: HResult;
      FExcepted: boolean;
      FExceptStr: string;
      FExceptionType: ExceptClass;
      FOwner: TgfglLiteWeightDownLoader;
      FHResultVal: HRESULT;
      FTriggeredDownloadingEvent: boolean;

      procedure SetHResultVal(const Value: HRESULT);
      procedure handleException(e: Exception);

   public

      procedure Init(Owner: TgfglLiteWeightDownLoader);

      function OnStartBinding(dwReserved: DWORD; pib: IBinding): HResult;
         stdcall;
      function GetPriority(out nPriority): HResult; stdcall;
      function OnLowResource(reserved: DWORD): HResult; stdcall;
      function OnProgress(ulProgress, ulProgressMax, ulStatusCode: ULONG;
         szStatusText: LPCWSTR): HResult; stdcall;
      function OnStopBinding(hresult: HResult; szError: LPCWSTR): HResult;
         stdcall;
      function GetBindInfo(out grfBINDF: DWORD; var bindinfo: TBindInfo):
      HResult; stdcall;
      function OnDataAvailable(grfBSCF: DWORD; dwSize: DWORD; formatetc:
         PFormatEtc;
         stgmed: PStgMedium): HResult; stdcall;
      function OnObjectAvailable(const iid: TGUID; punk: IUnknown): HResult;
         stdcall;

      property ErrorCode: HResult read FErrorCode;
      property Excepted: boolean read FExcepted;
      property ExceptStr: string read FExceptStr;
      property ExceptionType: ExceptClass read FExceptionType;
      property HResultVal: HRESULT read FHResultVal write SetHResultVal;

   end;

{ TUrlCallBack }

function TUrlCallBack.GetBindInfo(out grfBINDF: DWORD;
   var bindinfo: TBindInfo): HResult;
begin
   result := E_NOTIMPL;
//various experiments with caching
{    grfBINDF:=BINDF_RESYNCHRONIZE;
//   grfBINDF:=BINDF_RESYNCHRONIZE or or BINDF_GETNEWESTVERSION  or BINDF_PRAGMA_NO_CACHE  or BINDF_NOWRITECACHE  or BINDF_PULLDATA or BINDF_NEEDFILE ;
   sz:=bindinfo.cbSize;
   fillchar(bindinfo,sz,0);
   bindinfo.cbSize := sz;
   result := S_OK;
}
end;

function TUrlCallBack.GetPriority(out nPriority): HResult;
begin
   result := E_NOTIMPL;
end;

procedure TUrlCallBack.handleException(e: Exception);
begin
   FExcepted := true;
   FExceptStr := e.message;
   FExceptionType := ExceptClass(e.ClassType);
end;

procedure TUrlCallBack.Init(Owner: TgfglLiteWeightDownLoader);
begin
   FOwner := owner;
end;

function TUrlCallBack.OnDataAvailable(grfBSCF, dwSize: DWORD;
   formatetc: PFormatEtc; stgmed: PStgMedium): HResult;
begin
   result := E_NOTIMPL;
end;

function TUrlCallBack.OnLowResource(reserved: DWORD): HResult;
begin
   result := S_OK;
   try
      FOwner.doLowREsources;
   except
      on e: exception do begin
         handleException(e);
         result := E_FAIL;
      end;
   end;
end;

function TUrlCallBack.OnObjectAvailable(const iid: TGUID;
   punk: IUnknown): HResult;
begin
   result := E_NOTIMPL;
end;

function TUrlCallBack.OnProgress(ulProgress, ulProgressMax,
   ulStatusCode: ULONG; szStatusText: LPCWSTR): HResult;
var
   cancel: boolean;
begin
   result := s_OK;
   cancel := false;
   try
      try
         if (ulStatusCode = 5) and (not FTriggeredDownloadingEvent) then begin
            Fowner.doDownloading(ulProgressMax, cancel);
            FTriggeredDownloadingEvent := true;
            if cancel then EXIT;
         end;
         if ulStatusCode > ulong(high(TDownloadStatus)) then
            ulStatusCode := 0;
         FOwner.doProgress(ulprogress, ulProgressmax,
            TDownloadStatus(ulstatusCode),
            statusStrArray[TDownloadStatus(ulstatusCode)], string(szStatusText),
            cancel);
      except
         on e: exception do begin
            Cancel := true;
            handleException(e);
         end;
      end;
   finally
      if Cancel then
         result := E_ABORT;
   end;
end;

function TUrlCallBack.OnStartBinding(dwReserved: DWORD;
   pib: IBinding): HResult;
begin
   result := S_OK;
   try
      Fowner.doBegin;
   except
      on e: exception do begin
         handleException(e);
         result := E_INVALIDARG;
      end;
   end;
end;

function TUrlCallBack.OnStopBinding(hresult: HResult;
   szError: LPCWSTR): HResult;
begin
   result := S_OK;
   FHREsultVal := HResult;
   try
      FOwner.doEnded(HResult = E_ABORT, (HResult <> 1) and (HResult <> E_ABORT),
         FHREsultVal, sysErrorMessage(hresult))
   except
      on e: exception do begin
         handleException(e);
      end;
   end;
end;

procedure TUrlCallBack.SetHResultVal(const Value: HRESULT);
begin
   FHResultVal := Value;
end;

{ TgfglLiteWeightDownLoader }

procedure TgfglLiteWeightDownLoader.doEnded(aborted, error: boolean; errorCode:
   HResult; statusStr: string);
begin
   if assigned(FOnEnd) then
      FOnEnd(self, aborted, error, errorcode, statusStr);
end;

procedure TgfglLiteWeightDownLoader.doLowREsources;
begin
   if assigned(FOnLowResources) then
      FOnLowResources(self);
end;

procedure TgfglLiteWeightDownLoader.doProgress(position, max: integer;
   status: TDownLoadStatus; statusStr, extraInfoStr: string; var cancel:
   boolean);
begin
   if assigned(FOnProgress) then
      FOnProgress(self, position, max, status, statusStr, extraInfoStr, cancel);
end;

procedure TgfglLiteWeightDownLoader.doBegin;
begin
   if assigned(FonBegin) then
      FonBegin(self);
end;

{$HINTS off}

function TgfglLiteWeightDownLoader.Execute: boolean;
var
   httpFields: TStringlist;
   callback: TUrlCallBack;
   callbackI: IBindStatusCallBack;
   hr: HRESULT;
begin
   callback := nil;
   callbackI := nil;
   httpfields := nil;
   result := false;
   coinitialize(nil); //allow multithreading
   try
      try
         httpfields := TStringlist.create;
         callback := TUrlCallBack.create;
         callbackI := callback; //for avoiding premature free
         callback.init(self);

         ExtractHTTPFields(['/'], [], pchar(httpDecode(trim(FUrl))), httpFields,
            false);
         setlength(FFilePath, MAX_PATH + 1);
         if FFilePath = '' then begin
            hr := UrlDownloadToFile(nil,
               pchar(FUrl),
            {pchar(FFIlePath)}
               pchar(IncludeTrailingBackslash(trim(FDestinationFolder)) +
               httpFields[httpFields.count - 1]),
            {URLOSTRM_GETNEWESTVERSION ,}
               0,
               callBackI);
         end else begin
            hr := UrlDownloadToFile(nil,
               pchar(FUrl),
               pchar(FFilePath),
            {URLOSTRM_GETNEWESTVERSION ,}
               0,
               callBackI);
         end;

         if hr = E_OUTOFMEMORY then
            raise EWin32Error.Create(SysErrorMessage(hr));

         result := hr = s_ok;

         if callback.excepted then
            raise callback.ExceptionType.create(callback.FExceptStr);
      finally
         callbackI := nil;
      end;
   finally
      httpFields.free;
      couninitialize;
   end;
end;
{$HINTS on}

procedure TgfglLiteWeightDownLoader.SetUrl(const Value: string);
begin
   FUrl := Value;
end;

procedure TgfglLiteWeightDownLoader.doDownloading(size: integer;
   var cancel: boolean);
begin
   if assigned(FLWOnDownloading) then
      FLWOnDownloading(self, size, cancel);
end;

function extractObjectName(url: string): string;
var
   httpFields: TStringlist;
begin
   httpFields := TStringlist.create;
   try
      ExtractHTTPFields(['/'], [], pchar(httpDecode(trim(Url))), httpFields,
         false);
      result := httpFields[httpFields.count - 1]
   finally
      httpFields.Free;
   end;
end;

procedure Register;
begin
   registerComponents('custom', [TgfglLiteWeightDownLoader]);
end;

procedure TgfglLiteWeightDownLoader.SetFilePath(const Value: string);
begin
   FFilePath := Value;
end;

end.

